from typing import List

from Bio.motifs import create
from Bio.SeqIO import parse
from numpy.random import random
from tensorflow import float32, int64
from tensorflow.data import TFRecordDataset
from tensorflow.io import FixedLenFeature, TFRecordWriter, parse_single_example
from tensorflow.train import Example, Feature, Features, FloatList, Int64List

from constants import BASES, BATCH_SIZE, EPOCHS, TEST_DATASET, TRAIN_DATASET


def generate_example(sequence, weight_matrix) -> bytes:
    """
    Create a binary-string for each sequence containing the sequence and the bases' frequency
    """
    schema = {
        "sequence": Feature(
            int64_list=Int64List(value=list(encode_sequence(sequence)))
        ),
        "A_counts": Feature(float_list=FloatList(value=[weight_matrix["A"][0]])),
        "C_counts": Feature(float_list=FloatList(value=[weight_matrix["C"][0]])),
        "G_counts": Feature(float_list=FloatList(value=[weight_matrix["G"][0]])),
        "T_counts": Feature(float_list=FloatList(value=[weight_matrix["T"][0]])),
    }
    example = Example(features=Features(feature=schema))
    return example.SerializeToString()


def encode_sequence(sequence) -> List[int]:
    """
    Encode the DNA sequence using the indices of the BASES constant
    """
    encoded_sequence = [BASES.index(element) for element in sequence]
    return encoded_sequence


def read_fastq(filepath) -> List[bytes]:
    """
    Parse a FASTQ file and generate a List of serialized Examples
    """
    examples = []
    with open(filepath) as handle:
        for row in parse(handle, "fastq"):
            sequence = str(row.seq)
            motifs = create(row.seq)
            example = generate_example(sequence=sequence, weight_matrix=motifs.pwm)
            examples.append(example)
    return examples


def create_dataset(filepath) -> None:
    """
    Create a training and test dataset with a 70/30 split respectively
    """
    data = read_fastq(filepath)
    train_test_split = 0.7
    with TFRecordWriter(TRAIN_DATASET) as train, TFRecordWriter(TEST_DATASET) as test:
        for element in data:
            if random() < train_test_split:
                train.write(element)
            else:
                test.write(element)


def process_input(byte_string) -> Example:
    """
    Parse a byte-string into an Example object
    """
    schema = {
        "sequence": FixedLenFeature(shape=[], dtype=int64),
        "A_counts": FixedLenFeature(shape=[], dtype=float32),
        "C_counts": FixedLenFeature(shape=[], dtype=float32),
        "G_counts": FixedLenFeature(shape=[], dtype=float32),
        "T_counts": FixedLenFeature(shape=[], dtype=float32),
    }
    return parse_single_example(byte_string, features=schema)


def read_dataset() -> TFRecordDataset:
    """
    Read TFRecords files and generate a dataset
    """
    data_input = TFRecordDataset(filenames=[TRAIN_DATASET, TEST_DATASET])
    dataset = data_input.map(map_func=process_input)
    shuffled_dataset = dataset.shuffle(buffer_size=10000, reshuffle_each_iteration=True)
    batched_dataset = shuffled_dataset.batch(batch_size=BATCH_SIZE).repeat(count=EPOCHS)
    return batched_dataset
